import logging
from typing import Optional

from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.remote.remotexpc import RemoteXPCConnection


class RemoteService:
    def __init__(self, rsd: RemoteServiceDiscoveryService, service_name: str):
        self.service_name = service_name
        self.rsd = rsd
        self.service: Optional[RemoteXPCConnection] = None
        self.logger = logging.getLogger(self.__module__)

    def connect(self) -> None:
        self.service = self.rsd.start_remote_service(self.service_name)
        self.service.connect()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self) -> None:
        self.service.close()



from typing import List, Mapping, Optional

from pymobiledevice3.exceptions import ConnectionFailedError, PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService

MobileGestaltKeys = ['3GProximityCapability',
                     '3GVeniceCapability',
                     '3Gvenice',
                     '3d-imagery',
                     '3d-maps',
                     '64-bit',
                     '720p',
                     '720pPlaybackCapability',
                     'APNCapability',
                     'ARM64ExecutionCapability',
                     'ARMV6ExecutionCapability',
                     'ARMV7ExecutionCapability',
                     'ARMV7SExecutionCapability',
                     'ASTC',
                     'AWDID',
                     'AWDLCapability',
                     'AccelerometerCapability',
                     'AccessibilityCapability',
                     'AcousticID',
                     'ActivationProtocol',
                     'ActiveWirelessTechnology',
                     'ActuatorResonantFrequency',
                     'AdditionalTextTonesCapability',
                     'AggregateDevicePhotoZoomFactor',
                     'AggregateDeviceVideoZoomFactor',
                     'AirDropCapability',
                     'AirDropRestriction',
                     'AirplaneMode',
                     'AirplayMirroringCapability',
                     'AllDeviceCapabilities',
                     'Allow32BitApps',
                     'AllowOnlyATVCPSDKApps',
                     'AllowYouTube',
                     'AllowYouTubePlugin',
                     'AmbientLightSensorCapability',
                     'AmbientLightSensorSerialNumber',
                     'ApNonce',
                     'ApNonceRetrieve',
                     'AppCapacityTVOS',
                     'AppStore',
                     'AppStoreCapability',
                     'AppleInternalInstallCapability',
                     'AppleNeuralEngineSubtype',
                     'ApplicationInstallationCapability',
                     'ArcModuleSerialNumber',
                     'ArrowChipID',
                     'ArrowUniqueChipID',
                     'ArtworkTraits',
                     'AssistantCapability',
                     'AudioPlaybackCapability',
                     'AutoFocusCameraCapability',
                     'AvailableDisplayZoomSizes',
                     'BacklightCapability',
                     'BasebandAPTimeSync',
                     'BasebandBoardSnum',
                     'BasebandCertId',
                     'BasebandChipId',
                     'BasebandChipset',
                     'BasebandClass',
                     'BasebandFirmwareManifestData',
                     'BasebandFirmwareUpdateInfo',
                     'BasebandFirmwareVersion',
                     'BasebandKeyHashInformation',
                     'BasebandPostponementStatus',
                     'BasebandPostponementStatusBlob',
                     'BasebandRegionSKU',
                     'BasebandRegionSKURadioTechnology',
                     'BasebandSecurityInfoBlob',
                     'BasebandSerialNumber',
                     'BasebandSkeyId',
                     'BasebandStatus',
                     'BasebandUniqueId',
                     'BatteryCurrentCapacity',
                     'BatteryIsCharging',
                     'BatteryIsFullyCharged',
                     'BatterySerialNumber',
                     'BlueLightReductionSupported',
                     'BluetoothAddress',
                     'BluetoothAddressData',
                     'BluetoothCapability',
                     'BluetoothLE2Capability',
                     'BluetoothLECapability',
                     'BoardId',
                     'BoardRevision',
                     'BootManifestHash',
                     'BootNonce',
                     'BridgeBuild',
                     'BridgeRestoreVersion',
                     'BuddyLanguagesAnimationRequiresOptimization',
                     'BuildID',
                     'BuildVersion',
                     'C2KDeviceCapability',
                     'CPUArchitecture',
                     'CPUSubType',
                     'CPUType',
                     'CallForwardingCapability',
                     'CallWaitingCapability',
                     'CallerIDCapability',
                     'CameraAppUIVersion',
                     'CameraCapability',
                     'CameraFlashCapability',
                     'CameraFrontFlashCapability',
                     'CameraHDR2Capability',
                     'CameraHDRVersion',
                     'CameraLiveEffectsCapability',
                     'CameraMaxBurstLength',
                     'CameraRestriction',
                     'CarrierBundleInfoArray',
                     'CarrierInstallCapability',
                     'CellBroadcastCapability',
                     'CellularDataCapability',
                     'CellularTelephonyCapability',
                     'CertificateProductionStatus',
                     'CertificateSecurityMode',
                     'ChipID',
                     'CloudPhotoLibraryCapability',
                     'CoastlineGlowRenderingCapability',
                     'CompassCalibration',
                     'CompassCalibrationDictionary',
                     'CompassType',
                     'ComputerName',
                     'ConferenceCallType',
                     'ConfigNumber',
                     'ContainsCellularRadioCapability',
                     'ContinuityCapability',
                     'CoreRoutineCapability',
                     'CoverglassSerialNumber',
                     'DMin',
                     'DataPlanCapability',
                     'DebugBoardRevision',
                     'DelaySleepForHeadsetClickCapability',
                     'DesenseBuild',
                     'DeviceAlwaysPrewarmActuator',
                     'DeviceBackGlassMaterial',
                     'DeviceBackingColor',
                     'DeviceBrand',
                     'DeviceClass',
                     'DeviceClassNumber',
                     'DeviceColor',
                     'DeviceColorMapPolicy',
                     'DeviceCornerRadius',
                     'DeviceCoverGlassColor',
                     'DeviceCoverGlassMaterial',
                     'DeviceCoverMaterial',
                     'DeviceEnclosureColor',
                     'DeviceEnclosureMaterial',
                     'DeviceEnclosureRGBColor',
                     'DeviceHasAggregateCamera',
                     'DeviceHousingColor',
                     'DeviceIsMuseCapable',
                     'DeviceKeyboardCalibration',
                     'DeviceLaunchTimeLimitScale',
                     'DeviceName',
                     'DeviceNameString',
                     'DevicePrefers3DBuildingStrokes',
                     'DevicePrefersBuildingStrokes',
                     'DevicePrefersCheapTrafficShaders',
                     'DevicePrefersProceduralAntiAliasing',
                     'DevicePrefersTrafficAlpha',
                     'DeviceProximityCapability',
                     'DeviceRGBColor',
                     'DeviceRequiresPetalOptimization',
                     'DeviceRequiresProximityAmeliorations',
                     'DeviceRequiresSoftwareBrightnessCalculations',
                     'DeviceSceneUpdateTimeLimitScale',
                     'DeviceSubBrand',
                     'DeviceSupports1080p',
                     'DeviceSupports3DImagery',
                     'DeviceSupports3DMaps',
                     'DeviceSupports3rdPartyHaptics',
                     'DeviceSupports4G',
                     'DeviceSupports4k',
                     'DeviceSupports64Bit',
                     'DeviceSupports720p',
                     'DeviceSupports9Pin',
                     'DeviceSupportsAOP',
                     'DeviceSupportsARKit',
                     'DeviceSupportsASTC',
                     'DeviceSupportsAdaptiveMapsUI',
                     'DeviceSupportsAlwaysListening',
                     'DeviceSupportsAlwaysOnCompass',
                     'DeviceSupportsAlwaysOnTime',
                     'DeviceSupportsApplePencil',
                     'DeviceSupportsAutoLowLightVideo',
                     'DeviceSupportsAvatars',
                     'DeviceSupportsBatteryModuleAuthentication',
                     'DeviceSupportsBerkelium2',
                     'DeviceSupportsCCK',
                     'DeviceSupportsCameraCaptureOnTouchDown',
                     'DeviceSupportsCameraDeferredProcessing',
                     'DeviceSupportsCameraHaptics',
                     'DeviceSupportsCarIntegration',
                     'DeviceSupportsCinnamon',
                     'DeviceSupportsClosedLoopHaptics',
                     'DeviceSupportsCrudeProx',
                     'DeviceSupportsDClr',
                     'DeviceSupportsDoNotDisturbWhileDriving',
                     'DeviceSupportsELabel',
                     'DeviceSupportsEnhancedAC3',
                     'DeviceSupportsEnvironmentalDosimetry',
                     'DeviceSupportsExternalHDR',
                     'DeviceSupportsFloorCounting',
                     'DeviceSupportsHDRDeferredProcessing',
                     'DeviceSupportsHMEInARKit',
                     'DeviceSupportsHaptics',
                     'DeviceSupportsHardwareDetents',
                     'DeviceSupportsHeartHealthAlerts',
                     'DeviceSupportsHeartRateVariability',
                     'DeviceSupportsHiResBuildings',
                     'DeviceSupportsLineIn',
                     'DeviceSupportsLiquidDetection_CorrosionMitigation',
                     'DeviceSupportsLivePhotoAuto',
                     'DeviceSupportsLongFormAudio',
                     'DeviceSupportsMapsBlurredUI',
                     'DeviceSupportsMapsOpticalHeading',
                     'DeviceSupportsMomentCapture',
                     'DeviceSupportsNFC',
                     'DeviceSupportsNavigation',
                     'DeviceSupportsNewton',
                     'DeviceSupportsOnDemandPhotoAnalysis',
                     'DeviceSupportsP3ColorspaceVideoRecording',
                     'DeviceSupportsPeriodicALSUpdates',
                     'DeviceSupportsPhotosLocalLight',
                     'DeviceSupportsPortraitIntensityAdjustments',
                     'DeviceSupportsPortraitLightEffectFilters',
                     'DeviceSupportsRGB10',
                     'DeviceSupportsRaiseToSpeak',
                     'DeviceSupportsSiDP',
                     'DeviceSupportsSideButtonClickSpeed',
                     'DeviceSupportsSimplisticRoadMesh',
                     'DeviceSupportsSingleCameraPortrait',
                     'DeviceSupportsSiriBargeIn',
                     'DeviceSupportsSiriSpeaks',
                     'DeviceSupportsSiriSpokenMessages',
                     'DeviceSupportsSpatialOverCapture',
                     'DeviceSupportsStereoAudioRecording',
                     'DeviceSupportsStudioLightPortraitPreview',
                     'DeviceSupportsSwimmingWorkouts',
                     'DeviceSupportsTapToWake',
                     'DeviceSupportsTelephonyOverUSB',
                     'DeviceSupportsTethering',
                     'DeviceSupportsToneMapping',
                     'DeviceSupportsUSBTypeC',
                     'DeviceSupportsVSHCompensation',
                     'DeviceSupportsVoiceOverCanUseSiriVoice',
                     'DeviceSupportsWebkit',
                     'DeviceSupportsWirelessSplitting',
                     'DeviceSupportsYCbCr10',
                     'DeviceVariant',
                     'DeviceVariantGuess',
                     'DiagData',
                     'DictationCapability',
                     'DieId',
                     'DiskUsage',
                     'DisplayDriverICChipID',
                     'DisplayFCCLogosViaSoftwareCapability',
                     'DisplayMirroringCapability',
                     'DisplayPortCapability',
                     'DualSIMActivationPolicyCapable',
                     'EUICCChipID',
                     'EffectiveProductionStatus',
                     'EffectiveProductionStatusAp',
                     'EffectiveProductionStatusSEP',
                     'EffectiveSecurityMode',
                     'EffectiveSecurityModeAp',
                     'EffectiveSecurityModeSEP',
                     'EncodeAACCapability',
                     'EncryptedDataPartitionCapability',
                     'EnforceCameraShutterClick',
                     'EnforceGoogleMail',
                     'EthernetMacAddress',
                     'EthernetMacAddressData',
                     'ExplicitContentRestriction',
                     'ExternalChargeCapability',
                     'ExternalPowerSourceConnected',
                     'FDRSealingStatus',
                     'FMFAllowed',
                     'FaceTimeBackCameraTemporalNoiseReductionMode',
                     'FaceTimeBitRate2G',
                     'FaceTimeBitRate3G',
                     'FaceTimeBitRateLTE',
                     'FaceTimeBitRateWiFi',
                     'FaceTimeCameraRequiresFastSwitchOptions',
                     'FaceTimeCameraSupportsHardwareFaceDetection',
                     'FaceTimeDecodings',
                     'FaceTimeEncodings',
                     'FaceTimeFrontCameraTemporalNoiseReductionMode',
                     'FaceTimePhotosOptIn',
                     'FaceTimePreferredDecoding',
                     'FaceTimePreferredEncoding',
                     'FirmwareNonce',
                     'FirmwarePreflightInfo',
                     'FirmwareVersion',
                     'FirstPartyLaunchTimeLimitScale',
                     'ForwardCameraCapability',
                     'FrontCameraOffsetFromDisplayCenter',
                     'FrontCameraRotationFromDisplayNormal',
                     'FrontFacingCameraAutoHDRCapability',
                     'FrontFacingCameraBurstCapability',
                     'FrontFacingCameraCapability',
                     'FrontFacingCameraHDRCapability',
                     'FrontFacingCameraHDROnCapability',
                     'FrontFacingCameraHFRCapability',
                     'FrontFacingCameraHFRVideoCapture1080pMaxFPS',
                     'FrontFacingCameraHFRVideoCapture720pMaxFPS',
                     'FrontFacingCameraMaxVideoZoomFactor',
                     'FrontFacingCameraModuleSerialNumber',
                     'FrontFacingCameraStillDurationForBurst',
                     'FrontFacingCameraVideoCapture1080pMaxFPS',
                     'FrontFacingCameraVideoCapture4kMaxFPS',
                     'FrontFacingCameraVideoCapture720pMaxFPS',
                     'FrontFacingIRCameraModuleSerialNumber',
                     'FrontFacingIRStructuredLightProjectorModuleSerialNumber',
                     'Full6FeaturesCapability',
                     'GPSCapability',
                     'GSDeviceName',
                     'GameKitCapability',
                     'GasGaugeBatteryCapability',
                     'GreenTeaDeviceCapability',
                     'GyroscopeCapability',
                     'H264EncoderCapability',
                     'HDRImageCaptureCapability',
                     'HDVideoCaptureCapability',
                     'HEVCDecoder10bitSupported',
                     'HEVCDecoder12bitSupported',
                     'HEVCDecoder8bitSupported',
                     'HEVCEncodingCapability',
                     'HMERefreshRateInARKit',
                     'HWModelStr',
                     'HallEffectSensorCapability',
                     'HardwareEncodeSnapshotsCapability',
                     'HardwareKeyboardCapability',
                     'HardwarePlatform',
                     'HardwareSnapshotsRequirePurpleGfxCapability',
                     'HasAllFeaturesCapability',
                     'HasAppleNeuralEngine',
                     'HasBaseband',
                     'HasBattery',
                     'HasDaliMode',
                     'HasExtendedColorDisplay',
                     'HasIcefall',
                     'HasInternalSettingsBundle',
                     'HasMesa',
                     'HasPKA',
                     'HasSEP',
                     'HasSpringBoard',
                     'HasThinBezel',
                     'HealthKitCapability',
                     'HearingAidAudioEqualizationCapability',
                     'HearingAidLowEnergyAudioCapability',
                     'HearingAidPowerReductionCapability',
                     'HiDPICapability',
                     'HiccoughInterval',
                     'HideNonDefaultApplicationsCapability',
                     'HighestSupportedVideoMode',
                     'HomeButtonType',
                     'HomeScreenWallpaperCapability',
                     'IDAMCapability',
                     'IOSurfaceBackedImagesCapability',
                     'IOSurfaceFormatDictionary',
                     'IceFallID',
                     'IcefallInRestrictedMode',
                     'IcefallInfo',
                     'Image4CryptoHashMethod',
                     'Image4Supported',
                     'InDiagnosticsMode',
                     'IntegratedCircuitCardIdentifier',
                     'IntegratedCircuitCardIdentifier2',
                     'InternalBuild',
                     'InternationalMobileEquipmentIdentity',
                     'InternationalMobileEquipmentIdentity2',
                     'InternationalSettingsCapability',
                     'InverseDeviceID',
                     'IsEmulatedDevice',
                     'IsLargeFormatPhone',
                     'IsPwrOpposedVol',
                     'IsServicePart',
                     'IsSimulator',
                     'IsThereEnoughBatteryLevelForSoftwareUpdate',
                     'IsUIBuild',
                     'JasperSerialNumber',
                     'LTEDeviceCapability',
                     'LaunchTimeLimitScaleSupported',
                     'LisaCapability',
                     'LoadThumbnailsWhileScrollingCapability',
                     'LocalizedDeviceNameString',
                     'LocationRemindersCapability',
                     'LocationServicesCapability',
                     'LowPowerWalletMode',
                     'LunaFlexSerialNumber',
                     'LynxPublicKey',
                     'LynxSerialNumber',
                     'MLBSerialNumber',
                     'MLEHW',
                     'MMSCapability',
                     'MacBridgingKeys',
                     'MagnetometerCapability',
                     'MainDisplayRotation',
                     'MainScreenCanvasSizes',
                     'MainScreenClass',
                     'MainScreenHeight',
                     'MainScreenOrientation',
                     'MainScreenPitch',
                     'MainScreenScale',
                     'MainScreenStaticInfo',
                     'MainScreenWidth',
                     'MarketingNameString',
                     'MarketingProductName',
                     'MarketingVersion',
                     'MaxH264PlaybackLevel',
                     'MaximumScreenScale',
                     'MedusaFloatingLiveAppCapability',
                     'MedusaOverlayAppCapability',
                     'MedusaPIPCapability',
                     'MedusaPinnedAppCapability',
                     'MesaSerialNumber',
                     'MetalCapability',
                     'MicrophoneCapability',
                     'MicrophoneCount',
                     'MinimumSupportediTunesVersion',
                     'MixAndMatchPrevention',
                     'MobileDeviceMinimumVersion',
                     'MobileEquipmentIdentifier',
                     'MobileEquipmentInfoBaseId',
                     'MobileEquipmentInfoBaseProfile',
                     'MobileEquipmentInfoBaseVersion',
                     'MobileEquipmentInfoCSN',
                     'MobileEquipmentInfoDisplayCSN',
                     'MobileSubscriberCountryCode',
                     'MobileSubscriberNetworkCode',
                     'MobileWifi',
                     'ModelNumber',
                     'MonarchLowEndHardware',
                     'MultiLynxPublicKeyArray',
                     'MultiLynxSerialNumberArray',
                     'MultitaskingCapability',
                     'MultitaskingGesturesCapability',
                     'MusicStore',
                     'MusicStoreCapability',
                     'N78aHack',
                     'NFCRadio',
                     'NFCRadioCalibrationDataPresent',
                     'NFCUniqueChipID',
                     'NVRAMDictionary',
                     'NandControllerUID',
                     'NavajoFusingState',
                     'NikeIpodCapability',
                     'NotGreenTeaDeviceCapability',
                     'OLEDDisplay',
                     'OTAActivationCapability',
                     'OfflineDictationCapability',
                     'OpenGLES1Capability',
                     'OpenGLES2Capability',
                     'OpenGLES3Capability',
                     'OpenGLESVersion',
                     'PTPLargeFilesCapability',
                     'PanelSerialNumber',
                     'PanoramaCameraCapability',
                     'PartitionType',
                     'PasswordConfigured',
                     'PasswordProtected',
                     'PearlCameraCapability',
                     'PearlIDCapability',
                     'PeekUICapability',
                     'PeekUIWidth',
                     'Peer2PeerCapability',
                     'PersonalHotspotCapability',
                     'PhoneNumber',
                     'PhoneNumber2',
                     'PhosphorusCapability',
                     'PhotoAdjustmentsCapability',
                     'PhotoCapability',
                     'PhotoSharingCapability',
                     'PhotoStreamCapability',
                     'PhotosPostEffectsCapability',
                     'PiezoClickerCapability',
                     'PintoMacAddress',
                     'PintoMacAddressData',
                     'PipelinedStillImageProcessingCapability',
                     'PlatformStandAloneContactsCapability',
                     'PlatinumCapability',
                     'ProductHash',
                     'ProductName',
                     'ProductType',
                     'ProductVersion',
                     'ProximitySensorCalibration',
                     'ProximitySensorCalibrationDictionary',
                     'ProximitySensorCapability',
                     'RF-exposure-separation-distance',
                     'RFExposureSeparationDistance',
                     'RawPanelSerialNumber',
                     'RearCameraCapability',
                     'RearCameraOffsetFromDisplayCenter',
                     'RearFacingCamera60fpsVideoCaptureCapability',
                     'RearFacingCameraAutoHDRCapability',
                     'RearFacingCameraBurstCapability',
                     'RearFacingCameraCapability',
                     'RearFacingCameraHDRCapability',
                     'RearFacingCameraHDROnCapability',
                     'RearFacingCameraHFRCapability',
                     'RearFacingCameraHFRVideoCapture1080pMaxFPS',
                     'RearFacingCameraHFRVideoCapture720pMaxFPS',
                     'RearFacingCameraMaxVideoZoomFactor',
                     'RearFacingCameraModuleSerialNumber',
                     'RearFacingCameraStillDurationForBurst',
                     'RearFacingCameraSuperWideCameraCapability',
                     'RearFacingCameraTimeOfFlightCameraCapability',
                     'RearFacingCameraVideoCapture1080pMaxFPS',
                     'RearFacingCameraVideoCapture4kMaxFPS',
                     'RearFacingCameraVideoCapture720pMaxFPS',
                     'RearFacingCameraVideoCaptureFPS',
                     'RearFacingLowLightCameraCapability',
                     'RearFacingSuperWideCameraModuleSerialNumber',
                     'RearFacingTelephotoCameraCapability',
                     'RearFacingTelephotoCameraModuleSerialNumber',
                     'RecoveryOSVersion',
                     'RegionCode',
                     'RegionInfo',
                     'RegionSupportsCinnamon',
                     'RegionalBehaviorAll',
                     'RegionalBehaviorChinaBrick',
                     'RegionalBehaviorEUVolumeLimit',
                     'RegionalBehaviorGB18030',
                     'RegionalBehaviorGoogleMail',
                     'RegionalBehaviorNTSC',
                     'RegionalBehaviorNoPasscodeLocationTiles',
                     'RegionalBehaviorNoVOIP',
                     'RegionalBehaviorNoWiFi',
                     'RegionalBehaviorShutterClick',
                     'RegionalBehaviorValid',
                     'RegionalBehaviorVolumeLimit',
                     'RegulatoryModelNumber',
                     'ReleaseType',
                     'RemoteBluetoothAddress',
                     'RemoteBluetoothAddressData',
                     'RenderWideGamutImagesAtDisplayTime',
                     'RendersLetterPressSlowly',
                     'RequiredBatteryLevelForSoftwareUpdate',
                     'RestoreOSBuild',
                     'RestrictedCountryCodes',
                     'RingerSwitchCapability',
                     'RosalineSerialNumber',
                     'RoswellChipID',
                     'RotateToWakeStatus',
                     'SBAllowSensitiveUI',
                     'SBCanForceDebuggingInfo',
                     'SDIOManufacturerTuple',
                     'SDIOProductInfo',
                     'SEInfo',
                     'SEPNonce',
                     'SIMCapability',
                     'SIMPhonebookCapability',
                     'SIMStatus',
                     'SIMStatus2',
                     'SIMTrayStatus',
                     'SIMTrayStatus2',
                     'SMSCapability',
                     'SavageChipID',
                     'SavageInfo',
                     'SavageSerialNumber',
                     'SavageUID',
                     'ScreenDimensions',
                     'ScreenDimensionsCapability',
                     'ScreenRecorderCapability',
                     'ScreenSerialNumber',
                     'SecondaryBluetoothMacAddress',
                     'SecondaryBluetoothMacAddressData',
                     'SecondaryEthernetMacAddress',
                     'SecondaryEthernetMacAddressData',
                     'SecondaryWifiMacAddress',
                     'SecondaryWifiMacAddressData',
                     'SecureElement',
                     'SecureElementID',
                     'SecurityDomain',
                     'SensitiveUICapability',
                     'SerialNumber',
                     'ShoeboxCapability',
                     'ShouldHactivate',
                     'SiKACapability',
                     'SigningFuse',
                     'SiliconBringupBoard',
                     'SimultaneousCallAndDataCurrentlySupported',
                     'SimultaneousCallAndDataSupported',
                     'SiriGestureCapability',
                     'SiriOfflineCapability',
                     'Skey',
                     'SoftwareBehavior',
                     'SoftwareBundleVersion',
                     'SoftwareDimmingAlpha',
                     'SpeakerCalibrationMiGa',
                     'SpeakerCalibrationSpGa',
                     'SpeakerCalibrationSpTS',
                     'SphereCapability',
                     'StarkCapability',
                     'StockholmJcopInfo',
                     'StrictWakeKeyboardCases',
                     'SupportedDeviceFamilies',
                     'SupportedKeyboards',
                     'SupportsBurninMitigation',
                     'SupportsEDUMU',
                     'SupportsForceTouch',
                     'SupportsIrisCapture',
                     'SupportsLowPowerMode',
                     'SupportsPerseus',
                     'SupportsRotateToWake',
                     'SupportsSOS',
                     'SupportsSSHBButtonType',
                     'SupportsTouchRemote',
                     'SysCfg',
                     'SysCfgDict',
                     'SystemImageID',
                     'SystemTelephonyOfAnyKindCapability',
                     'TVOutCrossfadeCapability',
                     'TVOutSettingsCapability',
                     'TelephonyCapability',
                     'TelephonyMaximumGeneration',
                     'TimeSyncCapability',
                     'TopModuleAuthChipID',
                     'TouchDelivery120Hz',
                     'TouchIDCapability',
                     'TristarID',
                     'UIBackgroundQuality',
                     'UIParallaxCapability',
                     'UIProceduralWallpaperCapability',
                     'UIReachability',
                     'UMTSDeviceCapability',
                     'UnifiedIPodCapability',
                     'UniqueChipID',
                     'UniqueDeviceID',
                     'UniqueDeviceIDData',
                     'UserAssignedDeviceName',
                     'UserIntentPhysicalButtonCGRect',
                     'UserIntentPhysicalButtonCGRectString',
                     'UserIntentPhysicalButtonNormalizedCGRect',
                     'VOIPCapability',
                     'VeniceCapability',
                     'VibratorCapability',
                     'VideoCameraCapability',
                     'VideoStillsCapability',
                     'VoiceControlCapability',
                     'VolumeButtonCapability',
                     'WAGraphicQuality',
                     'WAPICapability',
                     'WLANBkgScanCache',
                     'WSKU',
                     'WatchCompanionCapability',
                     'WatchSupportsAutoPlaylistPlayback',
                     'WatchSupportsHighQualityClockFaceGraphics',
                     'WatchSupportsListeningOnGesture',
                     'WatchSupportsMusicStreaming',
                     'WatchSupportsSiriCommute',
                     'WiFiCallingCapability',
                     'WiFiCapability',
                     'WifiAddress',
                     'WifiAddressData',
                     'WifiAntennaSKUVersion',
                     'WifiCallingSecondaryDeviceCapability',
                     'WifiChipset',
                     'WifiFirmwareVersion',
                     'WifiVendor',
                     'WirelessBoardSnum',
                     'WirelessChargingCapability',
                     'YonkersChipID',
                     'YonkersSerialNumber',
                     'YonkersUID',
                     'YouTubeCapability',
                     'YouTubePluginCapability',
                     'accelerometer',
                     'accessibility',
                     'additional-text-tones',
                     'aggregate-cam-photo-zoom',
                     'aggregate-cam-video-zoom',
                     'airDropRestriction',
                     'airplay-mirroring',
                     'airplay-no-mirroring',
                     'all-features',
                     'allow-32bit-apps',
                     'ambient-light-sensor',
                     'ane',
                     'any-telephony',
                     'apn',
                     'apple-internal-install',
                     'applicationInstallation',
                     'arkit',
                     'arm64',
                     'armv6',
                     'armv7',
                     'armv7s',
                     'assistant',
                     'auto-focus',
                     'auto-focus-camera',
                     'baseband-chipset',
                     'bitrate-2g',
                     'bitrate-3g',
                     'bitrate-lte',
                     'bitrate-wifi',
                     'bluetooth',
                     'bluetooth-le',
                     'board-id',
                     'boot-manifest-hash',
                     'boot-nonce',
                     'builtin-mics',
                     'c2k-device',
                     'calibration',
                     'call-forwarding',
                     'call-waiting',
                     'caller-id',
                     'camera-flash',
                     'camera-front',
                     'camera-front-flash',
                     'camera-rear',
                     'cameraRestriction',
                     'car-integration',
                     'cell-broadcast',
                     'cellular-data',
                     'certificate-production-status',
                     'certificate-security-mode',
                     'chip-id',
                     'class',
                     'closed-loop',
                     'config-number',
                     'contains-cellular-radio',
                     'crypto-hash-method',
                     'dali-mode',
                     'data-plan',
                     'debug-board-revision',
                     'delay-sleep-for-headset-click',
                     'device-color-policy',
                     'device-colors',
                     'device-name',
                     'device-name-localized',
                     'dictation',
                     'die-id',
                     'display-mirroring',
                     'display-rotation',
                     'displayport',
                     'does-not-support-gamekit',
                     'effective-production-status',
                     'effective-production-status-ap',
                     'effective-production-status-sep',
                     'effective-security-mode',
                     'effective-security-mode-ap',
                     'effective-security-mode-sep',
                     'enc-top-type',
                     'encode-aac',
                     'encrypted-data-partition',
                     'enforce-googlemail',
                     'enforce-shutter-click',
                     'euicc-chip-id',
                     'explicitContentRestriction',
                     'face-detection-support',
                     'fast-switch-options',
                     'fcc-logos-via-software',
                     'fcm-type',
                     'firmware-version',
                     'flash',
                     'front-auto-hdr',
                     'front-burst',
                     'front-burst-image-duration',
                     'front-facing-camera',
                     'front-flash-capability',
                     'front-hdr',
                     'front-hdr-on',
                     'front-max-video-fps-1080p',
                     'front-max-video-fps-4k',
                     'front-max-video-fps-720p',
                     'front-max-video-zoom',
                     'front-slowmo',
                     'full-6',
                     'function-button_halleffect',
                     'function-button_ringerab',
                     'gamekit',
                     'gas-gauge-battery',
                     'gps',
                     'gps-capable',
                     'green-tea',
                     'gyroscope',
                     'h264-encoder',
                     'hall-effect-sensor',
                     'haptics',
                     'hardware-keyboard',
                     'has-sphere',
                     'hd-video-capture',
                     'hdr-image-capture',
                     'healthkit',
                     'hearingaid-audio-equalization',
                     'hearingaid-low-energy-audio',
                     'hearingaid-power-reduction',
                     'hiccough-interval',
                     'hide-non-default-apps',
                     'hidpi',
                     'home-button-type',
                     'homescreen-wallpaper',
                     'hw-encode-snapshots',
                     'hw-snapshots-need-purplegfx',
                     'iAP2Capability',
                     'iPadCapability',
                     'iTunesFamilyID',
                     'iap2-protocol-supported',
                     'image4-supported',
                     'international-settings',
                     'io-surface-backed-images',
                     'ipad',
                     'kConferenceCallType',
                     'kSimultaneousCallAndDataCurrentlySupported',
                     'kSimultaneousCallAndDataSupported',
                     'large-format-phone',
                     'live-effects',
                     'live-photo-capture',
                     'load-thumbnails-while-scrolling',
                     'location-reminders',
                     'location-services',
                     'low-power-wallet-mode',
                     'lte-device',
                     'magnetometer',
                     'main-screen-class',
                     'main-screen-height',
                     'main-screen-orientation',
                     'main-screen-pitch',
                     'main-screen-scale',
                     'main-screen-width',
                     'marketing-name',
                     'mesa',
                     'metal',
                     'microphone',
                     'mix-n-match-prevention-status',
                     'mms',
                     'modelIdentifier',
                     'multi-touch',
                     'multitasking',
                     'multitasking-gestures',
                     'n78a-mode',
                     'name',
                     'navigation',
                     'nfc',
                     'nfcWithRadio',
                     'nike-ipod',
                     'nike-support',
                     'no-coreroutine',
                     'no-hi-res-buildings',
                     'no-simplistic-road-mesh',
                     'not-green-tea',
                     'offline-dictation',
                     'opal',
                     'opengles-1',
                     'opengles-2',
                     'opengles-3',
                     'opposed-power-vol-buttons',
                     'ota-activation',
                     'panorama',
                     'peek-ui-width',
                     'peer-peer',
                     'personal-hotspot',
                     'photo-adjustments',
                     'photo-stream',
                     'piezo-clicker',
                     'pipelined-stillimage-capability',
                     'platinum',
                     'post-effects',
                     'pressure',
                     'prox-sensor',
                     'proximity-sensor',
                     'ptp-large-files',
                     'public-key-accelerator',
                     'rear-auto-hdr',
                     'rear-burst',
                     'rear-burst-image-duration',
                     'rear-cam-telephoto-capability',
                     'rear-facing-camera',
                     'rear-hdr',
                     'rear-hdr-on',
                     'rear-max-slomo-video-fps-1080p',
                     'rear-max-slomo-video-fps-720p',
                     'rear-max-video-fps-1080p',
                     'rear-max-video-fps-4k',
                     'rear-max-video-fps-720p',
                     'rear-max-video-frame_rate',
                     'rear-max-video-zoom',
                     'rear-slowmo',
                     'regulatory-model-number',
                     'ringer-switch',
                     'role',
                     's8000")',
                     's8003")',
                     'sandman-support',
                     'screen-dimensions',
                     'sensitive-ui',
                     'shoebox',
                     'sika-support',
                     'sim',
                     'sim-phonebook',
                     'siri-gesture',
                     'slow-letterpress-rendering',
                     'sms',
                     'software-bundle-version',
                     'software-dimming-alpha',
                     'stand-alone-contacts',
                     'still-camera',
                     'stockholm',
                     'supports-always-listening',
                     't7000")',
                     'telephony',
                     'telephony-maximum-generation',
                     'thin-bezel',
                     'tnr-mode-back',
                     'tnr-mode-front',
                     'touch-id',
                     'tv-out-crossfade',
                     'tv-out-settings',
                     'ui-background-quality',
                     'ui-no-parallax',
                     'ui-no-procedural-wallpaper',
                     'ui-pip',
                     'ui-reachability',
                     'ui-traffic-cheap-shaders',
                     'ui-weather-quality',
                     'umts-device',
                     'unified-ipod',
                     'unique-chip-id',
                     'venice',
                     'video-camera',
                     'video-cap',
                     'video-stills',
                     'voice-control',
                     'voip',
                     'volume-buttons',
                     'wapi',
                     'watch-companion',
                     'wi-fi',
                     'wifi',
                     'wifi-antenna-sku-info',
                     'wifi-chipset',
                     'wifi-module-sn',
                     'wlan',
                     'wlan.background-scan-cache',
                     'youtube',
                     'youtubePlugin']


class DiagnosticsService(LockdownService):
    """
    Provides an API to:
    * Query MobileGestalt & IORegistry keys.
    * Reboot, shutdown or put the device in sleep mode.
    """
    RSD_SERVICE_NAME = 'com.apple.mobile.diagnostics_relay.shim.remote'
    SERVICE_NAME = 'com.apple.mobile.diagnostics_relay'
    OLD_SERVICE_NAME = 'com.apple.iosdiagnostics.relay'

    def __init__(self, lockdown: LockdownServiceProvider):
        if isinstance(lockdown, LockdownClient):
            try:
                service = lockdown.start_lockdown_service(self.SERVICE_NAME)
                service_name = self.SERVICE_NAME
            except ConnectionFailedError:
                service = lockdown.start_lockdown_service(self.OLD_SERVICE_NAME)
                service_name = self.OLD_SERVICE_NAME
        else:
            service = None
            service_name = self.RSD_SERVICE_NAME

        super().__init__(lockdown, service_name, service=service)

    def mobilegestalt(self, keys: List[str] = None) -> Mapping:
        if keys is None or len(keys) == 0:
            keys = MobileGestaltKeys
        resp = self.service.send_recv_plist({'Request': 'MobileGestalt', 'MobileGestaltKeys': keys})

        if (resp['Status'] != 'Success') or (resp['Diagnostics']['MobileGestalt']['Status'] != 'Success'):
            raise PyMobileDevice3Exception('failed to query MobileGestalt')

        resp['Diagnostics']['MobileGestalt'].pop('Status')

        return resp['Diagnostics']['MobileGestalt']

    def action(self, action: str) -> Optional[Mapping]:
        response = self.service.send_recv_plist({'Request': action})
        if response['Status'] != 'Success':
            raise PyMobileDevice3Exception(f'failed to perform action: {action}')
        return response.get('Diagnostics')

    def restart(self):
        self.action('Restart')

    def shutdown(self):
        self.action('Shutdown')

    def sleep(self):
        self.action('Sleep')

    def info(self, diag_type: str = 'All') -> Mapping:
        return self.action(diag_type)

    def ioregistry(self, plane: str = None, name: str = None, ioclass: str = None):
        d = {}

        if plane:
            d['CurrentPlane'] = plane

        if name:
            d['EntryName'] = name

        if ioclass:
            d['EntryClass'] = ioclass

        d['Request'] = 'IORegistry'

        response = self.service.send_recv_plist(d)
        if response.get('Status') != 'Success':
            raise PyMobileDevice3Exception(f'got invalid response: {response}')

        dd = response.get('Diagnostics')
        if dd:
            return dd.get('IORegistry')
        return None

    def get_battery(self) -> Mapping:
        return self.ioregistry(ioclass='IOPMPowerSource')



import logging
from functools import update_wrapper
from pathlib import Path
from urllib.error import URLError

import click

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.exceptions import AlreadyMountedError, DeveloperDiskImageNotFoundError, NotMountedError, \
    UnsupportedCommandError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.mobile_image_mounter import DeveloperDiskImageMounter, MobileImageMounterService, \
    PersonalizedImageMounter, auto_mount

logger = logging.getLogger(__name__)


def catch_errors(func):
    def catch_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AlreadyMountedError:
            logger.error('Given image was already mounted')
        except UnsupportedCommandError:
            logger.error('Your iOS version doesn\'t support this command')

    return update_wrapper(catch_function, func)


@click.group()
def cli():
    """ mounter cli """
    pass


@cli.group()
def mounter():
    """ mounter options """
    pass


@mounter.command('list', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_list(service_provider: LockdownClient, color):
    """ list all mounted images """
    output = []

    images = MobileImageMounterService(lockdown=service_provider).copy_devices()
    for image in images:
        image_signature = image.get('ImageSignature')
        if image_signature is not None:
            image['ImageSignature'] = image_signature.hex()
        output.append(image)

    print_json(output, colored=color)


@mounter.command('lookup', cls=Command)
@click.option('--color/--no-color', default=True)
@click.argument('image_type')
def mounter_lookup(service_provider: LockdownClient, color, image_type):
    """ lookup mounter image type """
    try:
        signature = MobileImageMounterService(lockdown=service_provider).lookup_image(image_type)
        print_json(signature, colored=color)
    except NotMountedError:
        logger.error(f'Disk image of type: {image_type} is not mounted')


@mounter.command('umount-developer', cls=Command)
@catch_errors
def mounter_umount_developer(service_provider: LockdownClient):
    """ unmount Developer image """
    try:
        DeveloperDiskImageMounter(lockdown=service_provider).umount()
        logger.info('Developer image unmounted successfully')
    except NotMountedError:
        logger.error('Developer image isn\'t currently mounted')


@mounter.command('umount-personalized', cls=Command)
@catch_errors
def mounter_umount_personalized(service_provider: LockdownClient):
    """ unmount Personalized image """
    try:
        PersonalizedImageMounter(lockdown=service_provider).umount()
        logger.info('Personalized image unmounted successfully')
    except NotMountedError:
        logger.error('Personalized image isn\'t currently mounted')


@mounter.command('mount-developer', cls=Command)
@click.argument('image', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument('signature', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@catch_errors
def mounter_mount_developer(service_provider: LockdownClient, image: str, signature: str):
    """ mount developer image """
    DeveloperDiskImageMounter(lockdown=service_provider).mount(Path(image), Path(signature))
    logger.info('Developer image mounted successfully')


@mounter.command('mount-personalized', cls=Command)
@click.argument('image', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument('trust-cache', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument('build-manifest', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@catch_errors
def mounter_mount_personalized(service_provider: LockdownClient, image: str, trust_cache: str, build_manifest: str):
    """ mount personalized image """
    PersonalizedImageMounter(lockdown=service_provider).mount(Path(image), Path(build_manifest), Path(trust_cache))
    logger.info('Personalized image mounted successfully')


@mounter.command('auto-mount', cls=Command)
@click.option('-x', '--xcode', type=click.Path(exists=True, dir_okay=True, file_okay=False),
              help='Xcode application path used to figure out automatically the DeveloperDiskImage path')
@click.option('-v', '--version', help='use a different DeveloperDiskImage version from the one retrieved by lockdown'
                                      'connection')
def mounter_auto_mount(service_provider: LockdownServiceProvider, xcode: str, version: str):
    """ auto-detect correct DeveloperDiskImage and mount it """
    try:
        auto_mount(service_provider, xcode=xcode, version=version)
        logger.info('DeveloperDiskImage mounted successfully')
    except URLError:
        logger.warning('failed to query DeveloperDiskImage versions')
    except DeveloperDiskImageNotFoundError:
        logger.error('Unable to find the correct DeveloperDiskImage')
    except AlreadyMountedError:
        logger.error('DeveloperDiskImage already mounted')
    except PermissionError as e:
        logger.error(
            f'DeveloperDiskImage could not be saved to Xcode default path ({e.filename}). '
            f'Please make sure your user has the necessary permissions')


@mounter.command('query-developer-mode-status', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_query_developer_mode_status(service_provider: LockdownClient, color):
    """ Query developer mode status """
    print_json(MobileImageMounterService(lockdown=service_provider).query_developer_mode_status(), colored=color)


@mounter.command('query-nonce', cls=Command)
@click.option('--image-type')
@click.option('--color/--no-color', default=True)
def mounter_query_nonce(service_provider: LockdownClient, image_type: str, color: bool):
    """ Query nonce """
    print_json(MobileImageMounterService(lockdown=service_provider).query_nonce(image_type), colored=color)


@mounter.command('query-personalization-identifiers', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_query_personalization_identifiers(service_provider: LockdownClient, color):
    """ Query personalization identifiers """
    print_json(MobileImageMounterService(lockdown=service_provider).query_personalization_identifiers(), colored=color)


@mounter.command('query-personalization-manifest', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_query_personalization_manifest(service_provider: LockdownClient, color):
    """ Query personalization manifest """
    result = []
    mounter = MobileImageMounterService(lockdown=service_provider)
    for device in mounter.copy_devices():
        result.append(mounter.query_personalization_manifest(device['PersonalizedImageType'], device['ImageSignature']))
    print_json(result, colored=color)


@mounter.command('roll-personalization-nonce', cls=Command)
def mounter_roll_personalization_nonce(service_provider: LockdownClient):
    MobileImageMounterService(lockdown=service_provider).roll_personalization_nonce()


@mounter.command('roll-cryptex-nonce', cls=Command)
def mounter_roll_cryptex_nonce(service_provider: LockdownClient):
    """ Roll cryptex nonce (will reboot) """
    MobileImageMounterService(lockdown=service_provider).roll_cryptex_nonce()



import asyncio
import platform
import socket
import traceback
from functools import wraps
from typing import Callable

from construct import Int8ul, Int16ul, Int32ul, Int64ul, Select

DEFAULT_AFTER_IDLE_SEC = 3
DEFAULT_INTERVAL_SEC = 3
DEFAULT_MAX_FAILS = 3

_DARWIN_TCP_KEEPALIVE = 0x10
_DARWIN_TCP_KEEPINTVL = 0x101
_DARWIN_TCP_KEEPCNT = 0x102


def plist_access_path(d, path: tuple, type_=None, required=False):
    for component in path:
        d = d.get(component)
        if d is None:
            break

    if type_ == bool and isinstance(d, str):
        if d.lower() not in ('true', 'false'):
            raise ValueError()
        d = 'true' == d.lower()
    elif type_ is not None and not isinstance(d, type_):
        # wrong type
        d = None

    if d is None and required:
        raise KeyError(f'path: {path} doesn\'t exist in given plist object')

    return d


def bytes_to_uint(b: bytes):
    return Select(u64=Int64ul, u32=Int32ul, u16=Int16ul, u8=Int8ul).parse(b)


def try_decode(s: bytes):
    try:
        return s.decode('utf8')
    except UnicodeDecodeError:
        return s


def asyncio_print_traceback(f: Callable):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        try:
            return await f(*args, **kwargs)
        except Exception as e:  # noqa: E72
            if not isinstance(e, asyncio.CancelledError):
                traceback.print_exc()
            raise

    return wrapper


def set_keepalive(sock: socket.socket, after_idle_sec: int = DEFAULT_AFTER_IDLE_SEC,
                  interval_sec: int = DEFAULT_INTERVAL_SEC, max_fails: int = DEFAULT_MAX_FAILS) -> None:
    """
    set keep-alive parameters on a given socket

    :param sock: socket to operate on
    :param after_idle_sec: idle time used when SO_KEEPALIVE is enabled
    :param interval_sec: interval between keepalives
    :param max_fails: number of keepalives before close

    """
    plat = platform.system()
    if plat == 'Linux':
        return _set_keepalive_linux(sock, after_idle_sec, interval_sec, max_fails)
    if plat == 'Darwin':
        return _set_keepalive_darwin(sock, after_idle_sec, interval_sec, max_fails)
    if plat == 'Windows':
        return _set_keepalive_win(sock, after_idle_sec, interval_sec)
    raise RuntimeError(f'Unsupported platform {plat}')


def _set_keepalive_linux(sock: socket.socket, after_idle_sec: int = DEFAULT_AFTER_IDLE_SEC,
                         interval_sec: int = DEFAULT_INTERVAL_SEC, max_fails: int = DEFAULT_MAX_FAILS) -> None:
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, after_idle_sec)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, interval_sec)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, max_fails)


def _set_keepalive_darwin(sock: socket.socket, after_idle_sec: int = DEFAULT_AFTER_IDLE_SEC,
                          interval_sec: int = DEFAULT_INTERVAL_SEC, max_fails: int = DEFAULT_MAX_FAILS) -> None:
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
    sock.setsockopt(socket.IPPROTO_TCP, _DARWIN_TCP_KEEPALIVE, after_idle_sec)
    sock.setsockopt(socket.IPPROTO_TCP, _DARWIN_TCP_KEEPINTVL, interval_sec)
    sock.setsockopt(socket.IPPROTO_TCP, _DARWIN_TCP_KEEPCNT, max_fails)


def _set_keepalive_win(sock: socket.socket, after_idle_sec: int = DEFAULT_AFTER_IDLE_SEC,
                       interval_sec: int = DEFAULT_INTERVAL_SEC) -> None:
    sock.ioctl(socket.SIO_KEEPALIVE_VALS, (1, after_idle_sec * 1000, interval_sec * 1000))



import logging
from dataclasses import dataclass
from typing import List, Mapping, Optional, Tuple, Union

from pymobiledevice3.exceptions import InvalidServiceError, NoDeviceConnectedError, PyMobileDevice3Exception, \
    StartServiceError
from pymobiledevice3.lockdown import LockdownClient, create_using_remote
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.remote.bonjour import DEFAULT_BONJOUR_TIMEOUT, get_remoted_addresses
from pymobiledevice3.remote.remotexpc import RemoteXPCConnection
from pymobiledevice3.service_connection import LockdownServiceConnection


@dataclass
class RSDDevice:
    hostname: str
    udid: str
    product_type: str
    os_version: str


# from remoted ([RSDRemoteNCMDeviceDevice createPortListener])
RSD_PORT = 58783


class RemoteServiceDiscoveryService(LockdownServiceProvider):
    def __init__(self, address: Tuple[str, int]):
        super().__init__()
        self.service = RemoteXPCConnection(address)
        self.peer_info: Optional[Mapping] = None
        self.lockdown: Optional[LockdownClient] = None
        self.all_values: Optional[Mapping] = None
        self.all_domains: Optional[Mapping] = None

    @property
    def product_version(self) -> str:
        return self.peer_info['Properties']['OSVersion']

    @property
    def ecid(self) -> int:
        return self.peer_info['Properties']['UniqueChipID']

    @property
    def developer_mode_status(self) -> bool:
        return self.lockdown.developer_mode_status

    def connect(self) -> None:
        self.service.connect()
        self.peer_info = self.service.receive_response()
        self.udid = self.peer_info['Properties']['UniqueDeviceID']
        self.product_type = self.peer_info['Properties']['ProductType']
        try:
            self.lockdown = create_using_remote(self.start_lockdown_service('com.apple.mobile.lockdown.remote.trusted'))
        except InvalidServiceError:
            self.lockdown = create_using_remote(
                self.start_lockdown_service('com.apple.mobile.lockdown.remote.untrusted'))
        self.all_values = self.lockdown.all_values
        self.all_domains = self.lockdown.all_domains

    def get_value(self, domain: str = None, key: str = None):
        return self.lockdown.get_value(domain, key)

    def start_lockdown_service_without_checkin(self, name: str) -> LockdownServiceConnection:
        return LockdownServiceConnection.create_using_tcp(self.service.address[0], self.get_service_port(name))

    def start_lockdown_service(self, name: str, include_escrow_bag: bool = False) -> LockdownServiceConnection:
        service = self.start_lockdown_service_without_checkin(name)
        checkin = {'Label': 'pymobiledevice3', 'ProtocolVersion': '2', 'Request': 'RSDCheckin'}
        if include_escrow_bag:
            # TODO: read the `createRemoteUnlockKey` result
            raise NotImplementedError('EscrowBag over RemoteXPC is not yet supported')
        response = service.send_recv_plist(checkin)
        if response['Request'] != 'RSDCheckin':
            raise PyMobileDevice3Exception(f'Invalid response for RSDCheckIn: {response}. Expected "RSDCheckIn"')
        response = service.recv_plist()
        if response['Request'] != 'StartService':
            raise PyMobileDevice3Exception(f'Invalid response for RSDCheckIn: {response}. Expected "ServiceService"')
        return service

    async def aio_start_lockdown_service(
            self, name: str, include_escrow_bag: bool = False) -> LockdownServiceConnection:
        service = self.start_lockdown_service(name, include_escrow_bag=include_escrow_bag)
        await service.aio_start()
        return service

    def start_lockdown_developer_service(self, name, include_escrow_bag: bool = False) -> LockdownServiceConnection:
        try:
            return self.start_lockdown_service_without_checkin(name)
        except StartServiceError:
            logging.getLogger(self.__module__).error(
                'Failed to connect to required service. Make sure DeveloperDiskImage.dmg has been mounted. '
                'You can do so using: pymobiledevice3 mounter mount'
            )
            raise

    def start_remote_service(self, name: str) -> RemoteXPCConnection:
        service = RemoteXPCConnection((self.service.address[0], self.get_service_port(name)))
        return service

    def start_service(self, name: str) -> Union[RemoteXPCConnection, LockdownServiceConnection]:
        service = self.peer_info['Services'][name]
        service_properties = service.get('Properties', {})
        use_remote_xpc = service_properties.get('UsesRemoteXPC', False)
        return self.start_remote_service(name) if use_remote_xpc else self.start_lockdown_service(name)

    def get_service_port(self, name: str) -> int:
        """takes a service name and returns the port that service is running on if the service exists"""
        service = self.peer_info['Services'].get(name)
        if service is None:
            raise InvalidServiceError(f'No such service: {name}')
        return int(service['Port'])

    def close(self) -> None:
        self.service.close()
        if self.lockdown is not None:
            self.lockdown.close()

    def __enter__(self) -> 'RemoteServiceDiscoveryService':
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def __repr__(self) -> str:
        return (f'<{self.__class__.__name__} PRODUCT:{self.product_type} VERSION:{self.product_version} '
                f'UDID:{self.udid}>')


def get_remoted_devices(timeout: int = DEFAULT_BONJOUR_TIMEOUT) -> List[RSDDevice]:
    result = []
    for hostname in get_remoted_addresses(timeout):
        with RemoteServiceDiscoveryService((hostname, RSD_PORT)) as rsd:
            properties = rsd.peer_info['Properties']
            result.append(RSDDevice(hostname=hostname, udid=properties['UniqueDeviceID'],
                                    product_type=properties['ProductType'], os_version=properties['OSVersion']))
    return result


def get_remoted_device(udid: str, timeout: int = DEFAULT_BONJOUR_TIMEOUT) -> RSDDevice:
    devices = get_remoted_devices(timeout=timeout)
    for device in devices:
        if device.udid == udid:
            return device
    raise NoDeviceConnectedError()



import logging
import traceback

import click
import coloredlogs

from pymobiledevice3.exceptions import AccessDeniedError, ConnectionFailedToUsbmuxdError, DeveloperModeError, \
    DeveloperModeIsNotEnabledError, DeviceHasPasscodeSetError, DeviceNotFoundError, InternalError, \
    InvalidServiceError, MessageNotSupportedError, MissingValueError, NoDeviceConnectedError, NoDeviceSelectedError, \
    NotPairedError, PairingDialogResponsePendingError, PasswordRequiredError, SetProhibitedError, \
    TunneldConnectionError, UserDeniedPairingError

coloredlogs.install(level=logging.INFO)

logging.getLogger('quic').disabled = True
logging.getLogger('asyncio').disabled = True
logging.getLogger('zeroconf').disabled = True
logging.getLogger('parso.cache').disabled = True
logging.getLogger('parso.cache.pickle').disabled = True
logging.getLogger('parso.python.diff').disabled = True
logging.getLogger('humanfriendly.prompts').disabled = True
logging.getLogger('blib2to3.pgen2.driver').disabled = True
logging.getLogger('urllib3.connectionpool').disabled = True

logger = logging.getLogger(__name__)

INVALID_SERVICE_MESSAGE = """Failed to start service. Possible reasons are:
- If you were trying to access a developer service (developer subcommand):
    - Make sure the DeveloperDiskImage/PersonalizedImage is mounted via:
      > python3 -m pymobiledevice3 mounter auto-mount

    - If your device iOS version >= 17.0:
        - Make sure you passed the --rsd option to the subcommand
          https://github.com/doronz88/pymobiledevice3#working-with-developer-tools-ios--170

- Apple removed this service

- A bug. Please file a bug report:
  https://github.com/doronz88/pymobiledevice3/issues/new?assignees=&labels=&projects=&template=bug_report.md&title=
"""

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

# Mapping of index options to import file names
CLI_GROUPS = {
    'activation': 'activation',
    'afc': 'afc',
    'amfi': 'amfi',
    'apps': 'apps',
    'backup2': 'backup',
    'bonjour': 'bonjour',
    'companion': 'companion_proxy',
    'crash': 'crash',
    'developer': 'developer',
    'diagnostics': 'diagnostics',
    'lockdown': 'lockdown',
    'mounter': 'mounter',
    'notification': 'notification',
    'pcap': 'pcap',
    'power-assertion': 'power_assertion',
    'processes': 'processes',
    'profile': 'profile',
    'provision': 'provision',
    'remote': 'remote',
    'restore': 'restore',
    'springboard': 'springboard',
    'status': 'status',
    'syslog': 'syslog',
    'usbmux': 'usbmux',
    'webinspector': 'webinspector'
}


class Pmd3Cli(click.Group):
    def list_commands(self, ctx):
        return CLI_GROUPS.keys()

    def get_command(self, ctx, name):
        if name not in CLI_GROUPS.keys():
            ctx.fail(f'No such command {name!r}.')
        try:
            mod = __import__(f'pymobiledevice3.cli.{CLI_GROUPS[name]}', None, None, ['cli'])
        except ImportError:
            return
        command = mod.cli.get_command(ctx, name)
        # Some cli groups have different names than the index
        if not command:
            command_name = mod.cli.list_commands(ctx)[0]
            command = mod.cli.get_command(ctx, command_name)
        return command


@click.command(cls=Pmd3Cli, context_settings=CONTEXT_SETTINGS)
def cli():
    pass


def main() -> None:
    try:
        cli()
    except NoDeviceConnectedError:
        logger.error('Device is not connected')
    except ConnectionAbortedError:
        logger.error('Device was disconnected')
    except NotPairedError:
        logger.error('Device is not paired')
    except UserDeniedPairingError:
        logger.error('User refused to trust this computer')
    except PairingDialogResponsePendingError:
        logger.error('Waiting for user dialog approval')
    except SetProhibitedError:
        logger.error('lockdownd denied the access')
    except MissingValueError:
        logger.error('No such value')
    except DeviceHasPasscodeSetError:
        logger.error('Cannot enable developer-mode when passcode is set')
    except DeveloperModeError as e:
        logger.error(f'Failed to enable developer-mode. Error: {e}')
    except ConnectionFailedToUsbmuxdError:
        logger.error('Failed to connect to usbmuxd socket. Make sure it\'s running.')
    except MessageNotSupportedError:
        logger.error('Message not supported for this iOS version')
        traceback.print_exc()
    except InternalError:
        logger.error('Internal Error')
    except DeveloperModeIsNotEnabledError:
        logger.error('Developer Mode is disabled. You can try to enable it using: '
                     'python3 -m pymobiledevice3 amfi enable-developer-mode')
    except InvalidServiceError:
        logger.error(INVALID_SERVICE_MESSAGE)
    except NoDeviceSelectedError:
        return
    except PasswordRequiredError:
        logger.error('Device is password protected. Please unlock and retry')
    except AccessDeniedError:
        logger.error('This command requires root privileges. Consider retrying with "sudo".')
    except BrokenPipeError:
        traceback.print_exc()
    except TunneldConnectionError:
        logger.error(
            'Unable to connect to Tunneld. You can start one using:\n'
            'sudo python3 -m pymobiledevice3 remote tunneld')
    except DeviceNotFoundError as e:
        logger.error(f'Device not found: {e.udid}')


if __name__ == '__main__':
    main()



import logging

from pymobiledevice3.services.remote_server import MessageAux


class LocationSimulation:
    IDENTIFIER = 'com.apple.instruments.server.services.LocationSimulation'

    def __init__(self, dvt):
        self.logger = logging.getLogger(__name__)
        self._channel = dvt.make_channel(self.IDENTIFIER)

    def simulate_location(self, latitude: float, longitude: float) -> None:
        self._channel.simulateLocationWithLatitude_longitude_(MessageAux().append_obj(latitude).append_obj(longitude))
        self._channel.receive_plist()

    def stop(self) -> None:
        self._channel.stopLocationSimulation()



import asyncio
import dataclasses
import logging
import os
import signal
from contextlib import asynccontextmanager, suppress
from typing import Dict, Tuple

import fastapi
import ifaddr.netifaces
import uvicorn
import zeroconf
from fastapi import FastAPI
from packaging import version
from zeroconf import IPVersion
from zeroconf.asyncio import AsyncZeroconf

from pymobiledevice3.exceptions import InterfaceIndexNotFoundError
from pymobiledevice3.remote.common import TunnelProtocol
from pymobiledevice3.remote.module_imports import start_tunnel
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.remote.utils import stop_remoted

logger = logging.getLogger(__name__)

ZEROCONF_TIMEOUT = 3000
MIN_VERSION = '16.0.0'
UNINIT_ADDRESS = ('', 0)


@dataclasses.dataclass
class Tunnel:
    rsd: RemoteServiceDiscoveryService
    task: asyncio.Task = None
    address: Tuple[str, int] = UNINIT_ADDRESS


class TunneldCore:
    def __init__(self):
        self.adapters: Dict[int, str] = {}
        self.active_tunnels: Dict[int, Tunnel] = {}
        self._type = '_remoted._tcp.local.'
        self._name = 'ncm._remoted._tcp.local.'
        self._interval = .5
        self.tasks = []

    def start(self) -> None:
        """ Register all tasks """
        self.tasks = [
            asyncio.create_task(self.update_adapters(), name='update_adapters'),
            asyncio.create_task(self.remove_detached_devices(), name='remove_detached_devices'),
            asyncio.create_task(self.discover_new_devices(), name='discover_new_devices'),
        ]

    async def close(self):
        """ close all tasks """
        for task in self.tasks:
            task.cancel()
            with suppress(asyncio.CancelledError):
                await task

    def clear(self) -> None:
        """ Clear active tunnels """
        for udid, tunnel in self.active_tunnels.items():
            logger.info(f'Removing tunnel {tunnel.address}')
            tunnel.rsd.close()
            tunnel.task.cancel()
        self.active_tunnels = {}

    @staticmethod
    async def handle_new_tunnel(tun: Tunnel) -> None:
        """ Create new tunnel """
        async with start_tunnel(tun.rsd, protocol=TunnelProtocol.TCP) as tunnel_result:
            tun.address = tunnel_result.address, tunnel_result.port
            logger.info(f'Created tunnel --rsd {tun.address[0]} {tun.address[1]}')
            await tunnel_result.client.wait_closed()

    @staticmethod
    async def connect_rsd(address: str, port: int) -> RemoteServiceDiscoveryService:
        """ Connect to RSD """
        with stop_remoted():
            rsd = RemoteServiceDiscoveryService((address, port))
            rsd.connect()
            return rsd

    async def update_adapters(self) -> None:
        """ Constantly updates the 'adapters' dictionary with IPv6 addresses linked to network interfaces """
        while True:
            self.adapters = {iface.index: addr.ip[0] for iface in ifaddr.get_adapters() for addr in iface.ips if
                             addr.is_IPv6}
            await asyncio.sleep(self._interval)

    async def remove_detached_devices(self) -> None:
        """ Continuously checks if adapters were removed and removes associated tunnels """
        while True:
            # Find active tunnels that are no longer associated with adapters
            diff = list(set(self.active_tunnels.keys()) - set(self.adapters.keys()))
            # For each detached tunnel, cancel its task, log the removal, and remove it from the active tunnels
            for k in diff:
                self.active_tunnels[k].task.cancel()
                self.active_tunnels[k].rsd.close()
                logger.info(f'Removing tunnel {self.active_tunnels[k].address}')
                self.active_tunnels.pop(k)

            await asyncio.sleep(self._interval)

    def get_interface_index(self, address: str) -> int:
        """
        To address the issue of an unknown IPv6 scope id for a device, we employ a workaround.
        We maintain a mapping that associates the scope id with the adapter address.
        To resolve this, we remove the last segment (quartet) from both the adapter address and the target address.
        If there is a match, we retrieve the scope id associated with that adapter and use it.

        Disclaimer: Matching addresses based on their segments may result in interface collision in specific network
        configurations.
        """
        address_segments = address.split(':')[:-1]
        for k, v in self.adapters.items():
            if address_segments != v.split(':')[:-1]:
                continue
            return k
        raise InterfaceIndexNotFoundError(address=address)

    async def discover_new_devices(self) -> None:
        """ Continuously scans for devices advertising 'RSD' through IPv6 adapters """
        while True:
            # Search for devices advertising the specified service type and name
            async with AsyncZeroconf(ip_version=IPVersion.V6Only) as aiozc:
                try:
                    info = await aiozc.async_get_service_info(self._type, self._name, timeout=ZEROCONF_TIMEOUT)
                except zeroconf.Error as e:
                    logger.warning(e)
                    continue
                if info is None:
                    continue
                # Extract device details
                addr = info.parsed_addresses(IPVersion.V6Only)[0]
                try:
                    interface_index = self.get_interface_index(addr)
                except InterfaceIndexNotFoundError as e:
                    logger.warning(f'Failed to find interface index for {e.address}')
                    continue
                if interface_index in self.active_tunnels:
                    continue
                # Connect to the discovered device
                addr = f'{addr}%{interface_index}'
                try:
                    rsd = await self.connect_rsd(addr, info.port)
                except (TimeoutError, ConnectionError, OSError):
                    logger.warning(f'Failed to connect rsd to {addr}')
                    continue
                # Check unsupported devices with a product version below a minimum threshold
                if version.parse(rsd.product_version) < version.parse(MIN_VERSION):
                    logger.warning(f'{rsd.udid} Unsupported device {rsd.product_version} < {MIN_VERSION}')
                    continue
                logger.info(f'Creating tunnel for {addr}')
                tunnel = Tunnel(rsd)
                # Add the tunnel to the active tunnels and start a handling task
                tunnel.task = asyncio.create_task(self.handle_new_tunnel(tunnel))
                self.active_tunnels[interface_index] = tunnel
            await asyncio.sleep(self._interval)


class TunneldRunner:
    """ TunneldRunner orchestrate between the webserver and TunneldCore """

    @classmethod
    def create(cls, host: str, port: int) -> None:
        cls(host, port)._run_app()

    def __init__(self, host: str, port: int):
        @asynccontextmanager
        async def lifespan(app: FastAPI):
            logging.getLogger('zeroconf').disabled = True
            self._tunneld_core.start()
            yield
            logger.info('Closing tunneld tasks...')
            await self._tunneld_core.close()

        self.host = host
        self.port = port
        self._app = FastAPI(lifespan=lifespan)
        self._tunneld_core = TunneldCore()

        @self._app.get('/')
        async def list_tunnels() -> Dict[str, Tuple[str, int]]:
            """ Retrieve the available tunnels and format them as {UUID: TUNNEL_ADDRESS} """
            tunnels = {}
            for k, v in self._tunneld_core.active_tunnels.items():
                if v.address == UNINIT_ADDRESS:
                    continue
                tunnels[v.rsd.udid] = v.address
            return tunnels

        @self._app.get('/shutdown')
        async def shutdown() -> fastapi.Response:
            """ Shutdown Tunneld """
            os.kill(os.getpid(), signal.SIGINT)
            return fastapi.Response(status_code=200, content='Server shutting down...')

        @self._app.get('/clear_tunnels')
        async def clear_tunnels() -> fastapi.Response:
            self._tunneld_core.clear()
            return fastapi.Response(status_code=200, content='Cleared tunnels...')

    def _run_app(self) -> None:
        uvicorn.run(self._app, host=self.host, port=self.port, loop='asyncio')



import logging
import plistlib

import click

from pymobiledevice3.cli.cli_common import Command, CommandWithoutAutopair, print_json
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.heartbeat import HeartbeatService

logger = logging.getLogger(__name__)


@click.group()
def cli():
    """ apps cli """
    pass


@cli.group('lockdown')
def lockdown_group():
    """ lockdown options """
    pass


@lockdown_group.command('recovery', cls=Command)
def lockdown_recovery(service_provider: LockdownClient):
    """ enter recovery """
    print_json(service_provider.enter_recovery())


@lockdown_group.command('service', cls=Command)
@click.argument('service_name')
def lockdown_service(service_provider: LockdownServiceProvider, service_name):
    """ send-receive raw service messages with a given service name"""
    service_provider.start_lockdown_service(service_name).shell()


@lockdown_group.command('developer-service', cls=Command)
@click.argument('service_name')
def lockdown_developer_service(service_provider: LockdownServiceProvider, service_name):
    """ send-receive raw service messages with a given developer service name """
    service_provider.start_lockdown_developer_service(service_name).shell()


@lockdown_group.command('info', cls=Command)
@click.option('-a', '--all', is_flag=True, help='include all domain information')
@click.option('--color/--no-color', default=True)
def lockdown_info(service_provider: LockdownServiceProvider, all, color):
    """ query all lockdown values """
    print_json(service_provider.all_domains if all else service_provider.all_values, colored=color)


@lockdown_group.command('get', cls=Command)
@click.argument('domain', required=False)
@click.argument('key', required=False)
@click.option('--color/--no-color', default=True)
def lockdown_get(service_provider: LockdownClient, domain, key, color):
    """ query lockdown values by their domain and key names """
    print_json(service_provider.get_value(domain=domain, key=key), colored=color)


@lockdown_group.command('set', cls=Command)
@click.argument('value')
@click.argument('domain', required=False)
@click.argument('key', required=False)
@click.option('--color/--no-color', default=True)
def lockdown_set(service_provider: LockdownClient, value, domain, key, color):
    """ set a lockdown value using python's eval() """
    print_json(service_provider.set_value(value=eval(value), domain=domain, key=key), colored=color)


@lockdown_group.command('remove', cls=Command)
@click.argument('domain')
@click.argument('key')
@click.option('--color/--no-color', default=True)
def lockdown_remove(service_provider: LockdownClient, domain, key, color):
    """ remove a domain/key pair """
    print_json(service_provider.remove_value(domain=domain, key=key), colored=color)


@lockdown_group.command('unpair', cls=CommandWithoutAutopair)
@click.argument('host_id', required=False)
def lockdown_unpair(service_provider: LockdownClient, host_id: str = None):
    """ unpair from connected device """
    service_provider.unpair(host_id=host_id)


@lockdown_group.command('pair', cls=CommandWithoutAutopair)
def lockdown_pair(service_provider: LockdownClient):
    """ pair device """
    service_provider.pair()


@lockdown_group.command('save-pair-record', cls=CommandWithoutAutopair)
@click.argument('output', type=click.File('wb'))
def lockdown_save_pair_record(service_provider: LockdownClient, output):
    """ save pair record to specified location """
    if service_provider.pair_record is None:
        logger.error('no pairing record was found')
        return
    plistlib.dump(service_provider.pair_record, output)


@lockdown_group.command('date', cls=Command)
def lockdown_date(service_provider: LockdownClient):
    """ get device date """
    print(service_provider.date)


@lockdown_group.command('heartbeat', cls=Command)
def lockdown_heartbeat(service_provider: LockdownClient):
    """ start heartbeat service """
    HeartbeatService(service_provider).start()


@lockdown_group.command('language', cls=Command)
def lockdown_language(service_provider: LockdownClient):
    """ get current language settings """
    print(f'{service_provider.language} {service_provider.locale}')


@lockdown_group.command('device-name', cls=Command)
@click.argument('new_name', required=False)
def lockdown_device_name(service_provider: LockdownClient, new_name):
    """ get/set current device name """
    if new_name:
        service_provider.set_value(new_name, key='DeviceName')
    else:
        print(f'{service_provider.get_value(key="DeviceName")}')


@lockdown_group.command('wifi-connections', cls=Command)
@click.argument('state', type=click.Choice(['on', 'off']), required=False)
def lockdown_wifi_connections(service_provider: LockdownClient, state):
    """ get/set wifi connections state """
    if not state:
        # show current state
        print_json(service_provider.get_value(domain='com.apple.mobile.wireless_lockdown'))
    else:
        # enable/disable
        service_provider.enable_wifi_connections = state == 'on'



import asyncio
import logging
import sys
import tempfile
from functools import partial
from typing import List, TextIO

import click

from pymobiledevice3.cli.cli_common import RSDCommand, print_json, prompt_device_list, sudo_required
from pymobiledevice3.exceptions import NoDeviceConnectedError
from pymobiledevice3.remote.bonjour import get_remoted_addresses
from pymobiledevice3.remote.common import TunnelProtocol
from pymobiledevice3.remote.module_imports import MAX_IDLE_TIMEOUT, start_tunnel, verify_tunnel_imports
from pymobiledevice3.remote.remote_service_discovery import RSD_PORT, RemoteServiceDiscoveryService
from pymobiledevice3.remote.utils import TUNNELD_DEFAULT_ADDRESS, stop_remoted
from pymobiledevice3.tunneld import TunneldRunner

logger = logging.getLogger(__name__)


def get_device_list() -> List[RemoteServiceDiscoveryService]:
    result = []
    with stop_remoted():
        for address in get_remoted_addresses():
            rsd = RemoteServiceDiscoveryService((address, RSD_PORT))
            try:
                rsd.connect()
            except ConnectionRefusedError:
                continue
            result.append(rsd)
    return result


@click.group()
def cli():
    """ remote cli """
    pass


@cli.group('remote')
def remote_cli():
    """ remote options """
    pass


@remote_cli.command('tunneld')
@click.option('--host', default=TUNNELD_DEFAULT_ADDRESS[0])
@click.option('--port', type=click.INT, default=TUNNELD_DEFAULT_ADDRESS[1])
@click.option('-d', '--daemonize', is_flag=True)
@sudo_required
def cli_tunneld(host: str, port: int, daemonize: bool):
    """ Start Tunneld service for remote tunneling """
    if not verify_tunnel_imports():
        return
    tunneld_runner = partial(TunneldRunner.create, host, port)
    if daemonize:
        try:
            from daemonize import Daemonize
        except ImportError:
            raise NotImplementedError('daemonizing is only supported on unix platforms')
        with tempfile.NamedTemporaryFile('wt') as pid_file:
            daemon = Daemonize(app=f'Tunneld {host}:{port}', pid=pid_file.name,
                               action=tunneld_runner)
            logger.info(f'starting Tunneld {host}:{port}')
            daemon.start()
    else:
        tunneld_runner()


@remote_cli.command('browse')
@click.option('--color/--no-color', default=True)
def browse(color: bool):
    """ browse devices using bonjour """
    devices = []
    for rsd in get_device_list():
        devices.append({'address': rsd.service.address[0],
                        'port': RSD_PORT,
                        'UniqueDeviceID': rsd.peer_info['Properties']['UniqueDeviceID'],
                        'ProductType': rsd.peer_info['Properties']['ProductType'],
                        'OSVersion': rsd.peer_info['Properties']['OSVersion']})
    print_json(devices, colored=color)


@remote_cli.command('rsd-info', cls=RSDCommand)
@click.option('--color/--no-color', default=True)
def rsd_info(service_provider: RemoteServiceDiscoveryService, color: bool):
    """ show info extracted from RSD peer """
    print_json(service_provider.peer_info, colored=color)


async def tunnel_task(
        service_provider: RemoteServiceDiscoveryService, secrets: TextIO,
        script_mode: bool = False, max_idle_timeout: float = MAX_IDLE_TIMEOUT,
        protocol: TunnelProtocol = TunnelProtocol.QUIC) -> None:
    if start_tunnel is None:
        raise NotImplementedError('failed to start the QUIC tunnel on your platform')

    async with start_tunnel(service_provider, secrets=secrets, max_idle_timeout=max_idle_timeout,
                            protocol=protocol) as tunnel_result:
        logger.info('tunnel created')
        if script_mode:
            print(f'{tunnel_result.address} {tunnel_result.port}')
        else:
            if secrets is not None:
                print(click.style('Secrets: ', bold=True, fg='magenta') +
                      click.style(secrets.name, bold=True, fg='white'))
            print(click.style('UDID: ', bold=True, fg='yellow') +
                  click.style(service_provider.udid, bold=True, fg='white'))
            print(click.style('ProductType: ', bold=True, fg='yellow') +
                  click.style(service_provider.product_type, bold=True, fg='white'))
            print(click.style('ProductVersion: ', bold=True, fg='yellow') +
                  click.style(service_provider.product_version, bold=True, fg='white'))
            print(click.style('Interface: ', bold=True, fg='yellow') +
                  click.style(tunnel_result.interface, bold=True, fg='white'))
            print(click.style('Protocol: ', bold=True, fg='yellow') +
                  click.style(tunnel_result.protocol, bold=True, fg='white'))
            print(click.style('RSD Address: ', bold=True, fg='yellow') +
                  click.style(tunnel_result.address, bold=True, fg='white'))
            print(click.style('RSD Port: ', bold=True, fg='yellow') +
                  click.style(tunnel_result.port, bold=True, fg='white'))
            print(click.style('Use the follow connection option:\n', bold=True, fg='yellow') +
                  click.style(f'--rsd {tunnel_result.address} {tunnel_result.port}', bold=True, fg='cyan'))
        sys.stdout.flush()
        await tunnel_result.client.wait_closed()
        logger.info('tunnel was closed')


@remote_cli.command('start-tunnel')
@click.option('--udid', help='UDID for a specific device to look for')
@click.option('--secrets', type=click.File('wt'), help='TLS keyfile for decrypting with Wireshark')
@click.option('--script-mode', is_flag=True,
              help='Show only HOST and port number to allow easy parsing from external shell scripts')
@click.option('--max-idle-timeout', type=click.FLOAT, default=MAX_IDLE_TIMEOUT,
              help='Maximum QUIC idle time (ping interval)')
@click.option('-p', '--protocol', type=click.Choice([e.value for e in TunnelProtocol]),
              default=TunnelProtocol.TCP.value)
@sudo_required
def cli_start_tunnel(udid: str, secrets: TextIO, script_mode: bool, max_idle_timeout: float, protocol: str):
    """ start quic tunnel """
    protocol = TunnelProtocol(protocol)
    if not verify_tunnel_imports():
        return
    devices = get_device_list()
    if not devices:
        # no devices were found
        raise NoDeviceConnectedError()
    if len(devices) == 1:
        # only one device found
        rsd = devices[0]
    else:
        # several devices were found
        if udid is None:
            # show prompt if non explicitly selected
            rsd = prompt_device_list(devices)
        else:
            rsd = [device for device in devices if device.udid == udid]
            if len(rsd) > 0:
                rsd = rsd[0]
            else:
                raise NoDeviceConnectedError()

    if udid is not None and rsd.udid != udid:
        raise NoDeviceConnectedError()

    asyncio.run(tunnel_task(rsd, secrets, script_mode, max_idle_timeout=max_idle_timeout, protocol=protocol),
                debug=True)


@remote_cli.command('service', cls=RSDCommand)
@click.argument('service_name')
def cli_service(service_provider: RemoteServiceDiscoveryService, service_name: str):
    """ start an ipython shell for interacting with given service """
    with service_provider.start_remote_service(service_name) as service:
        service.shell()
